import collections
import re
import string
import numpy as np
from collections import Iterable

ERR_NA = '#N/A'
ERR_VALUE = '#VALUE!'
ERR_REF = '#REF!'
XLFN = '_xlfn.'
ERR_NAME = '#NAME?'
ERR_DIV = '#DIV/0!'

def get_range(spreadsheet, cell_address):
    return spreadsheet.cellmap[cell_address].address()

def evaluate_spreadsheet(spreadsheet, sheet, col_nr, row_nr):
    return spreadsheet.evaluate(sheet + '!' + col_nr + str(row_nr))

def to_values(spreadsheet, rng, skip_none=True, is_valid_function=None):
    result = []
    def process(col_nr, row_nr, value):
        if (not skip_none or value is not None) and is_valid_function(value):
            result.append(value)
        return True
    iterate_cells(spreadsheet, rng, process)
    return result


"""
nrg can be a single range string or a tuple of ranges.
"""
def iterate_cells(spreadsheet, rng, process_fn):
    if isinstance(rng, list):
        for rg in rng:
            iterate_cells_single(spreadsheet, rg, process_fn)
    else:
        iterate_cells_single(spreadsheet, rng, process_fn)

def iterate_cells_single(spreadsheet, rng, process_fn):
    sheet, start, end = split_range(rng)
    sh, start_col, start_row = split_address(start)
    sh, end_col, end_row = split_address(end)
    start_col_nr, end_col_nr = col2num(start_col), col2num(end_col)
    start_row_nr, end_row_nr = int(start_row), int(end_row)
    for row_nr in range(start_row_nr, end_row_nr + 1):
        for col_nr in range(start_col_nr, end_col_nr + 1):
            if (not process_fn(col_nr, row_nr, evaluate_spreadsheet(spreadsheet, sheet, num2col(col_nr), row_nr))):
                return

def rankdata(a, method='average'):
    """
    Assign ranks to data, dealing with ties appropriately.
    Ranks begin at 1.  The `method` argument controls how ranks are assigned
    to equal values.  See [1]_ for further discussion of ranking methods.
    Parameters
    ----------
    a : array_like
        The array of values to be ranked.  The array is first flattened.
    method : str, optional
        The method used to assign ranks to tied elements.
        The options are 'average', 'min', 'max', 'dense' and 'ordinal'.
        'average':
            The average of the ranks that would have been assigned to
            all the tied values is assigned to each value.
        'min':
            The minimum of the ranks that would have been assigned to all
            the tied values is assigned to each value.  (This is also
            referred to as "competition" ranking.)
        'max':
            The maximum of the ranks that would have been assigned to all
            the tied values is assigned to each value.
        'dense':
            Like 'min', but the rank of the next highest element is assigned
            the rank immediately after those assigned to the tied elements.
        'ordinal':
            All values are given a distinct rank, corresponding to the order
            that the values occur in `a`.
        The default is 'average'.
    Returns
    -------
    ranks : ndarray
         An array of length equal to the size of `a`, containing rank
         scores.
    References
    ----------
    .. [1] "Ranking", http://en.wikipedia.org/wiki/Ranking
    .. [2] https://github.com/scipy/scipy/blob/master/scipy/stats/stats.py
    Examples
    --------
    """
    if method not in ('average', 'min', 'max', 'dense', 'ordinal'):
        raise ValueError('unknown method "{0}"'.format(method))

    arr = np.ravel(np.asarray(a))
    algo = 'mergesort' if method == 'ordinal' else 'quicksort'
    sorter = np.argsort(arr, kind=algo)

    inv = np.empty(sorter.size, dtype=np.intp)
    inv[sorter] = np.arange(sorter.size, dtype=np.intp)

    if method == 'ordinal':
        return inv + 1

    arr = arr[sorter]
    obs = np.r_[True, arr[1:] != arr[:-1]]
    dense = obs.cumsum()[inv]

    if method == 'dense':
        return dense

    # cumulative counts of each unique value
    count = np.r_[np.nonzero(obs)[0], len(obs)]

    if method == 'max':
        return count[dense]

    if method == 'min':
        return count[dense - 1] + 1

    # average method
    return .5 * (count[dense] + count[dense - 1] + 1)

def is_eval_allowed(self, cell):
    return not isinstance(cell.compiled_expression, str)
    # isinstance(cell.compiled_expression, str) and re.match('[A-Za-z]+', cell.compiled_expression)

def is_compile_allowed(self, obj):
    return obj.formula is not None

    # This doesn't work in case the value is a string that is the result of a formula like a vlookup, as such that the value
    # will be copied to the compiled expression and used, such that updated will not occur.
    # return not isinstance(obj.value, str)
    #
    # isinstance(self.value, str) and re.match('[A-Za-z]+', self.value):

def create_string_from_cell_value(value):
    # If given value is a String, checks if value contains multiple lines.
    # If value does contain multiple lines, value is split into lines and
    # rebuild to a single String with \\n appended between every line.
    # If value is not a String, just the quotes are added if not already present
    # and the result is returned.
    if not isinstance(value, str):
        raise TypeError('Value is not a String')
    lines = value.splitlines()
    if len(lines) > 1:
        result = '\\n'.join(lines)
    else:
        result = value
    if not result.startswith('"'):
        result = '"' + result
    if not result.endswith('"'):
        result = result + '"'
    return result

def clean_formula(formula):
    return remove_xlfn(formula)

def remove_xlfn(formula):
    return formula.replace(XLFN, '', 1) if formula else formula

def correct_range(excel, rng):
    return correct_range_max(rng, excel.max_columns(), excel.max_rows())

def correct_range_max(rng, max_col, max_row):
    sheet, start, end = split_range(rng)
    sh, start_col, start_row = split_address(start)
    sh, end_col, end_row = split_address(end)

    start_col_nr = col2num(start_col)
    end_col_nr = col2num(end_col)
    if (start_col_nr > end_col_nr):
        raise ValueError("Incorrect range specified: [%s], max_col: [%s], max_row: [%s]" % (rng, max_col, max_row))
    mx_col_start = min(start_col_nr, max_col)
    mx_col_end = min(end_col_nr, max_col)

    if start_row:
        start_row_nr = int(start_row)
        mx_row_start = min(start_row_nr, max_row)
    else:
        mx_row_start = 1

    if end_row:
        end_row_nr = int(end_row)
        mx_row_end = min(end_row_nr, max_row)
    else:
        mx_row_end = max_row

    if mx_col_start < start_col_nr or mx_col_end < end_col_nr or start_row is None or mx_row_start < start_row_nr or end_row is None or mx_row_end < end_row_nr:
        # We are requesting more cell and available in the excel, let's correct it
        result = num2col(mx_col_start) if start_col_nr > mx_col_start  else str(start_col)
        result += str(mx_row_start) + ':'
        result += num2col(mx_col_end) if end_col_nr > mx_col_end  else str(end_col)
        result += str(mx_row_end)

        #        result =  mx_col_start + str(mx_row_start) + ':' + num2col(mx_col_end) + str(mx_row_end)
        return sheet + '!' + result if sheet else result

    return rng

class CellRange(object):
    def __init__(self, address, value, sheet=None):
        self.__address = address.replace('$', '')
        sh, start, end = split_range(address)
        if not sh and not sheet:
            raise Exception("Must pass in a sheet")

        # make sure the address is always prefixed with the range
        if sh:
            sheet = sh
        else:
            self.__address = sheet + "!" + self.__address

        addr, nrows, ncols = resolve_range(address, sheet=sheet)

        # don't allow messing with these params
        self.__celladdr = addr
        self.__nrows = nrows
        self.__ncols = ncols
        self.__sheet = sheet
        self.value = value
        # We assume that a new cell range isn't dirty as it's value will be set
        self.__dirty = False
        self.__eval_inprogress = False
        self.__reset_inprogress = False

    def __str__(self):
        return self.__address

    def address(self):
        return self.__address

    """
    Used to indicate the cell range is dirty, which means re-evaluation is required
    """
    @property
    def dirty(self):
        return self.__dirty

    @dirty.setter
    def dirty(self, value):
        self.__dirty = value

    """
    Used to indicate the cell is being evaluated
    """
    @property
    def eval_inprogress(self):
        return self.__eval_inprogress

    @eval_inprogress.setter
    def eval_inprogress(self, value):
        self.__eval_inprogress = value

    """
    Used to indicate the cell is being reset
    """
    @property
    def reset_inprogress(self):
        return self.__reset_inprogress

    @reset_inprogress.setter
    def reset_inprogress(self, value):
        self.__reset_inprogress = value

    @property
    def celladdrs(self):
        return self.__celladdr

    @property
    def nrows(self):
        return self.__nrows

    @property
    def ncols(self):
        return self.__ncols

    @property
    def sheet(self):
        return self.__sheet

class Cell(object):
    ctr = 0
    # __dirty = False

    @classmethod
    def next_id(cls):
        cls.ctr += 1
        return cls.ctr

    def __init__(self, address, sheet, formula, value):
        super(Cell, self).__init__()

        # remove $'s
        address = address.replace('$', '')

        sh, c, r = split_address(address)

        # both are empty
        if not sheet and not sh:
            raise Exception("Sheet name may not be empty for cell address %s" % address)
        # both exist but disagree
        elif sh and sheet and sh != sheet:
            raise Exception("Sheet name mismatch for cell address %s: %s vs %s" % (address, sheet, sh))
        elif not sh and sheet:
            sh = sheet
        else:
            pass

        # we assume a cell's location can never change
        # __ Means that it is unchangeable
        self.__sheet = str(sheet)
        self.__sheet = sh
        self.__col = c
        self.__row = int(r) if r else None
        self.__col_idx = col2num(c)

        # Checks if formula is None, if so set formula to None

        if formula:
            self.__formula = formula
            self.value = value
        elif formula is None:
            self.__formula = None
            self.value = value
        elif isinstance(self.value, str):
            self.value = value

        # We assume that a new cell isn't dirty as it's value is set above
        self.__dirty = False
        #        self.__dirty = self.value is None
        self.__eval_inprogress = False
        self.__reset_inprogress = False

        self.python_expression = None
        self._compiled_expression = None

        # every cell has a unique id
        self.__id = Cell.next_id()

    """
    Used to indicate the cell is dirty, which means re-evaluation is required
    """
    @property
    def dirty(self):
        return self.__dirty

    @dirty.setter
    def dirty(self, value):
        self.__dirty = value

    """
    Used to indicate the cell is being evaluated
    """
    @property
    def eval_inprogress(self):
        return self.__eval_inprogress

    @eval_inprogress.setter
    def eval_inprogress(self, value):
        self.__eval_inprogress = value

    """
    Used to indicate the cell is being reset
    """
    @property
    def reset_inprogress(self):
        return self.__reset_inprogress

    @reset_inprogress.setter
    def reset_inprogress(self, value):
        self.__reset_inprogress = value

    @property
    def sheet(self):
        return self.__sheet

    @property
    def row(self):
        return self.__row

    @property
    def col(self):
        return self.__col

    @property
    def formula(self):
        return self.__formula

    @property
    def id(self):
        return self.__id

    @property
    def compiled_expression(self):
        return self._compiled_expression

    # code objects are not serializable
    def __getstate__(self):
        d = dict(self.__dict__)
        f = '_compiled_expression'
        if f in d:
            del d[f]
        return d

    def __setstate__(self, d):
        self.__dict__.update(d)
        self.compile()

    def clean_name(self):
        return self.address().replace('!', '_').replace(' ', '_')

    def address(self, absolute=True):
        if absolute:
            return "%s!%s%s" % (self.__sheet, self.__col, self.__row)
        else:
            return "%s%s" % (self.__col, self.__row)

    def address_parts(self):
        return self.__sheet, self.__col, self.__row, self.__col_idx

    def compile(self):
        if not self.python_expression:
            return

        # if we are a constant string, surround by quotes
        if (isinstance(self.value, str) or isinstance(self.value, int)) \
                and self.formula and self.python_expression.startswith('"'):
            self.python_expression = '"' + self.python_expression + '"'

        try:
            if is_compile_allowed(self, self):
                self._compiled_expression = compile(self.python_expression, '<string>', 'eval')
            else:
                self._compiled_expression = self.value
        except Exception as e:
            raise Exception(
                "Failed to compile cell %s with expression %s: %s" % (self.address(), self.python_expression, e))

    def __str__(self):
        if self.formula:
            return "%s%s" % (self.address(), self.formula)
        else:
            return "%s=%s" % (self.address(), self.value)

    @staticmethod
    def inc_col_address(address, inc):
        sh, col, row = split_address(address)
        return "%s!%s%s" % (sh, num2col(col2num(col) + inc), row)

    @staticmethod
    def inc_row_address(address, inc):
        sh, col, row = split_address(address)
        return "%s!%s%s" % (sh, col, row + inc)

    @staticmethod
    def resolve_cell(excel, address, sheet=None):
        r = excel.get_range(address)
        f = r.Formula if r.Formula.startswith('=') else None
        v = r.value

        sh, c, r = split_address(address)

        # use the sheet specified in the cell, else the passed sheet
        if sh:
            sheet = sh

        c = Cell(address, sheet, value=v, formula=f)
        return c

    @staticmethod
    def make_cells_multi(excel: object, ranges: object) -> object:
        result = []
        for sheet, rng in ranges.items():
            # starting points
            cursheet = sheet if sheet else excel.get_active_sheet()
            excel.set_sheet(cursheet)

            cells, nrows, ncols, rng = Cell.make_cells(excel, rng, cursheet)
            result.extend(list(flatten(cells)))
        return result

    @staticmethod
    def make_cells(excel: object, rng: object, sheet=None) -> object:
        """
        :rtype: object
        """
        cells = []

        def convert_range(rng, sheet=None):
            cells = []

            # use the sheet specified in the range, else the passed sheet
            sh, start, end = split_range(rng)
            if sh:
                sheet = sh

            # TODOEd: what is still required in this method in case of a range ? As the range
            # is directly supplied to the function (see issue in Redmine for more details)
            # Correct the range if its larger then the available cells in the sheet
            rng = correct_range(excel, rng)

            # get everything in blocks, is faster
            r = excel.get_range(rng)
            fs = r.Formula
            vs = r.value
            # print("Address, Formula and Value of range: [%s], fs: [%s], vs: [%s]:" % (rng, fs, vs))

            ads, numrows, numcols = resolve_range(rng)
            # ensure in the same nested format as fs/vs will be
            if numrows == 1:
                ads = [ads]
            elif numcols == 1:
                ads = [[x] for x in ads]

            # In case the value is a single integer, we must make it a string else the zip will throw an error indicating
            # it's not iteratable.
            # In case vs is None we have to wrap it in an array as well, else the zip function below will fail as None
            # isn't iteratable
            if vs is None or isinstance(vs, (int, float, str)):
                vs = [[str(vs)]]

            # We put the formula in a list, as a single string might result in problems.
            # Encountered problem was: ads=[['AV5']], fs='=AV178' that resulted below in the zip in fs='='
            # Appearantly unzipping (zip(*x)) will use the "=" as some kind of separator.
            if fs is None or isinstance(fs, str):
                fs = [[fs]]

            print("Creating cells for range: [%s]" % (rng))
            try:
                # We can't merge ads, fs and vs as vs can be less that can result in cells being ignored
                # Example: last column contains only formulas such that vs returns rows with 1 column less then a fs row
                for ad_row_nr in range(len(ads)):
                    row = []
                    ad_row = ads[ad_row_nr]
                    for ad_cell_nr in range(len(ad_row)):
                        a = ad_row[ad_cell_nr]
                        f = clean_formula(fs[ad_row_nr][ad_cell_nr])
                        f = f if f and f.startswith('=') else None
                        v = vs[ad_row_nr]
                        # A value can be empty, example: the last column contains only formulas
                        v = v[ad_cell_nr] if ad_cell_nr < len(v) else None
                        cl = Cell(a, sheet, value=v, formula=f)
                        # We expect a formula that is at least 2 character long, the first character should be the "="
                        if f is not None and len(f) < 2:
                            raise Exception(
                                "Wrong formula encountered: [%s], a: [%s], v: [%s], ads: [%s], fs: [%s], "
                                "vs: [%s], range: [%s], cell: [%s]" % (f, a, v, ads, fs, vs, rng, cl))
                        row.append(cl)
                        print("Created cell: [%s]" % (cl))
                    if (len(row) > 0):
                        cells.append(row)

                # return as vector
                if numrows == 1 or numcols == 1:
                    if len(cells) < 1:
                        raise Exception("No cells present in range, num_rows: [%s], num_cols: [%s], range: [%s]"
                                        % (numrows, numcols, rng))
                    if numrows == 1:
                        cells = cells[0]
                    else:
                        cls = []
                        for x in cells:
                            # We ignore empty cells, like cells that result in an excel #REF error.
                            if x is not None:
                                cls.append(x[0] if isinstance(x, list) else x)

                                # Removed: as it creates an error for empty/removed (#REF excel error) cells
                                # cells = [x[0] for x in cells]
                        cells = cls
                return cells, numrows, numcols, rng
            except Exception as ex:
                raise Exception(
                    "Error creating row cells for range [%s], sheet: [%s], ads: [%s], num_rows: [%s], num_cols: [%s], formula: [%s],"
                    " values: [%s], cells: [%s]" % (rng, sheet, ads, numrows, numcols, fs, vs, cells)) from ex

        def create_range_exception(rng, cell, sheet):
            return create_exception(rng, cell, sheet, "Error converting")

        def create_resolve_exception(rng, cell, sheet):
            return create_exception(rng, cell, sheet, "Error resolving")

        def create_exception(rng, cell, sheet, msg):
            return Exception(msg + ", range: [%s], cell: [%s], sheet: [%s], excel: [%s]"
                             % (rng, cell, sheet, excel))

        if isinstance(rng, list):  # if a list of cells
            cells = []
            for cell in rng:
                if is_range(cell):
                    try:
                        cs_in_range, nr, nc, rng = convert_range(cell, sheet)
                        cells.append(cs_in_range)
                    except Exception as ex:
                        raise create_range_exception(rng, cell, sheet) from ex
                else:
                    try:
                        c = Cell.resolve_cell(excel, cell, sheet=sheet)
                        cells.append(c)
                    except Exception as ex:
                        raise create_resolve_exception(rng, cell, sheet) from ex

            cells = list(flatten(cells))
            return cells, -1, -1, None  # numrows, numcols and range are irrelevant here, so we return nr=nc=-1
        else:
            if is_range(rng):
                try:
                    # We return the range as it will be specific in case the shorthand range notitation was used,
                    # shorthand: (A:B), and specific: (A88:B99)
                    cells, numrows, numcols, rng = convert_range(rng, sheet)
                except Exception as ex:
                    raise create_range_exception(rng, None, sheet) from ex
            else:
                try:
                    c = Cell.resolve_cell(excel, rng, sheet=sheet)
                    cells.append(c)
                    numrows = 1
                    numcols = 1
                except Exception as ex:
                    raise create_resolve_exception(rng, None, sheet) from ex

        return cells, numrows, numcols, rng


        # @classmethod
        # def resolve_cell(cls, excel, dep, sheet):
        #     pass

def is_range(address):
    return address.find(':') > 0

def split_range(rng):
    if rng.find('!') > 0:
        sh, r = rng.split("!")
        start, end = r.split(':')
    else:
        sh = None
        start, end = rng.split(':')

    return sh, start, end

def create_sheet(rng):
    if rng.find('!') > 0:
        sh, r = rng.split("!")
        return sh
    return None

def split_address(address):
    sheet = None
    if address.find('!') > 0:
        sheet, address = address.split('!')

    # ignore case
    address = address.upper()

    # regular <col><row> format    
    if re.match('^[A-Z $]+[\d $]+$', address):
        col, row = [_f for _f in re.split('([A-Z $]+)', address) if _f]
    # shorthand range notition (A:H) instead of (A1:H100) (al rows in A till all rows in H)
    elif re.match('^[A-Z $]+$', address):
        col = address
        row = None
    # R<row>C<col> format
    elif re.match('^R\d+C\d+$', address):
        row, col = address.split('C')
        row = row[1:]
    # R[<row>]C[<col>] format
    elif re.match('^R\[\d+]C\[\d+]$', address):
        row, col = address.split('C')
        row = row[2:-1]
        col = col[2:-1]
    else:
        raise Exception('Invalid address format ' + address)

    return sheet, col, row

def resolve_range(rng, flatten=False, sheet=''):
    sh, start, end = split_range(rng)

    if sh and sheet:
        if sh != sheet:
            raise Exception("Mismatched sheets %s and %s" % (sh, sheet))
        else:
            sheet += '!'
    elif sh and not sheet:
        sheet = sh + "!"
    elif sheet and not sh:
        sheet += "!"
    else:
        pass

    # single cell, no range
    if not is_range(rng):
        return [sheet + rng], 1, 1

    sh, start_col, start_row = split_address(start)
    sh, end_col, end_row = split_address(end)
    start_col_idx = col2num(start_col)
    end_col_idx = col2num(end_col)

    if start_row and not end_row or not start_row and end_row:
        raise Exception(
            "One row is missing (both must be (not) present, start row: " + start_row + ", end row: " + end_row)

    # # Let's return some "marker" values we use to recognize the shorthand notition when required.
    # if start_row is None and end_row is None:
    #     return 'all', -1, -1

    start_row = int(start_row)
    end_row = int(end_row)

    # single column
    if start_col == end_col:
        nrows = end_row - start_row + 1
        data = ["%s%s%s" % (s, c, r) for (s, c, r) in
                zip([sheet] * nrows, [start_col] * nrows, list(range(start_row, end_row + 1)))]
        return data, len(data), 1

    # single row
    elif start_row == end_row:
        ncols = end_col_idx - start_col_idx + 1
        data = ["%s%s%s" % (s, num2col(c), r) for (s, c, r) in
                zip([sheet] * ncols, list(range(start_col_idx, end_col_idx + 1)), [start_row] * ncols)]
        return data, 1, len(data)

    # rectangular range
    else:
        cells = []
        for r in range(start_row, end_row + 1):
            row = []
            for c in range(start_col_idx, end_col_idx + 1):
                row.append(sheet + num2col(c) + str(r))

            cells.append(row)

        if flatten:
            # flatten into one list
            l = flatten(cells)
            return l, 1, len(l)
        else:
            return cells, len(cells), len(cells[0])

            # e.g., convert BA -> 53

def col2num(col):
    if not col:
        raise Exception("Column may not be empty")

    tot = 0
    for i, c in enumerate([c for c in col[::-1] if c != "$"]):
        if c == '$':
            continue
        tot += (ord(c) - 64) * 26 ** i
    return tot

# convert back
def num2col(num):
    if num < 1:
        raise Exception("Number must be larger than 0: %s" % num)

    s = ''
    q = num
    while q > 0:
        (q, r) = divmod(q, 26)
        if r == 0:
            q = q - 1
            r = 26
        s = string.ascii_uppercase[r - 1] + s
    return s

def address2index(a):
    sh, c, r = split_address(a)
    return col2num(c), int(r)

def index2addres(c, r, sheet=None):
    return "%s%s%s" % (sheet + "!" if sheet else "", num2col(c), r)

def get_linest_degree(excel, cl):
    # TODO: assumes a row or column of linest formulas & that all coefficients are needed

    sh, c, r, ci = cl.address_parts()
    # figure out where we are in the row

    # to the left
    i = ci - 1
    while i > 0:
        f = excel.get_formula_from_range(index2addres(i, r))
        if f is None or f != cl.formula:
            break
        else:
            i = i - 1

    # to the right
    j = ci + 1
    while True:
        f = excel.get_formula_from_range(index2addres(j, r))
        if f is None or f != cl.formula:
            break
        else:
            j = j + 1

    # assume the degree is the number of linest's
    degree = (j - i - 1) - 1  # last -1 is because an n degree polynomial has n+1 coefs

    # which coef are we (left most coef is the coef for the highest power)
    coef = ci - i

    # no linests left or right, try looking up/down
    if degree == 0:
        # up
        i = r - 1
        while i > 0:
            f = excel.get_formula_from_range("%s%s" % (c, i))
            if f is None or f != cl.formula:
                break
            else:
                i = i - 1

        # down
        j = r + 1
        while True:
            f = excel.get_formula_from_range("%s%s" % (c, j))
            if f is None or f != cl.formula:
                break
            else:
                j = j + 1

        degree = (j - i - 1) - 1
        coef = r - i

    # if degree is zero -> only one linest formula -> linear regression -> degree should be one
    return max(degree, 1), coef

def flatten(l):
    for el in l:
        if isinstance(el, collections.Iterable) and not isinstance(el, str):
            for sub in flatten(el):
                yield sub
        else:
            yield el

def uniqueify(seq):
    seen = set()
    seen_add = seen.add
    return [x for x in seq if x not in seen and not seen_add(x)]


def is_valid(value, criteria):
    check = criteria_parser(criteria)
    return check(value)

def is_collection(args):
    return isinstance(args, (list, tuple))

def is_nr(value):
    return value is not None and isinstance(value, (int, float))


#TODOMathijs: should we use this, or the above?
def is_number(s):  # http://stackoverflow.com/questions/354038/how-do-i-check-if-a-string-is-a-number-float-in-python
    try:
        float(s)
        return True
    except ValueError:
        return False

def is_leap_year(year):
    if not is_number(year):
        raise TypeError("%s must be a number" % str(year))
    if year <= 0:
        raise TypeError("%s must be strictly positive" % str(year))

    # Watch out, 1900 is a leap according to Excel => https://support.microsoft.com/en-us/kb/214326
    return (year % 4 == 0 and year % 100 != 0 or year % 400 == 0) or year == 1900

def get_max_days_in_month(month, year):
    if not is_number(year) or not is_number(month):
        raise TypeError("All inputs must be a number")
    if year <= 0 or month <= 0:
        raise TypeError("All inputs must be strictly positive")

    if month in (4, 6, 9, 11):
        return 30
    elif month == 2:
        if is_leap_year(year):
            return 29
        else:
            return 28
    else:
        return 31

def normalize_year(y, m, d):
    if m <= 0:
        y -= int(abs(m) / 12 + 1)
        m = 12 - (abs(m) % 12)
        normalize_year(y, m, d)
    elif m > 12:
        y += int(m / 12)
        m = m % 12

    if d <= 0:
        d += get_max_days_in_month(m, y)
        m -= 1
        y, m, d = normalize_year(y, m, d)

    else:
        if m in (4, 6, 9, 11) and d > 30:
            m += 1
            d -= 30
            y, m, d = normalize_year(y, m, d)
        elif m == 2:
            if (is_leap_year(y)) and d > 29:
                m += 1
                d -= 29
                y, m, d = normalize_year(y, m, d)
            elif (not is_leap_year(y)) and d > 28:
                m += 1
                d -= 28
                y, m, d = normalize_year(y, m, d)
        elif d > 31:
            m += 1
            d -= 31
            y, m, d = normalize_year(y, m, d)

    return y, m, d

def date_from_int(nb):
    if not is_number(nb):
        raise TypeError("%s is not a number" % str(nb))

    # origin of the Excel date system
    current_year = 1900
    current_month = 0
    current_day = 0

    while (nb > 0):
        if not is_leap_year(current_year) and nb > 365:
            current_year += 1
            nb -= 365
        elif is_leap_year(current_year) and nb > 366:
            current_year += 1
            nb -= 366
        else:
            current_month += 1
            max_days = get_max_days_in_month(current_month, current_year)

            if nb > max_days:
                nb -= max_days
            else:
                current_day = nb
                nb = 0

    return current_year, current_month, current_day

def criteria_parser(criteria):
    if is_number(criteria):
        def check(x):
            return x == criteria  # and type(x) == type(criteria)
    elif type(criteria) == str:
        search = re.search('(\W*)(.*)', criteria.lower()).group
        operator = search(1)
        value = search(2)
        value = float(value) if is_number(value) else str(value)

        if operator == '<':
            def check(x):
                if not is_number(x):
                    return False; # Added such that it skips non-numbers in for example the sumif function
                    # raise TypeError(
                    #     'excellib.countif() doesnt\'t work for checking non number items against non equality')
                return x < value
        elif operator == '>':
            def check(x):
                if not is_number(x):
                    return False;
                    # raise TypeError(
                    #     'excellib.countif() doesnt\'t work for checking non number items against non equality')
                return x > value
        elif operator == '>=':
            def check(x):
                print(('\n TEST', x))
                if not is_number(x):
                    return False;
                    # raise TypeError(
                    #     'excellib.countif() doesnt\'t work for checking non number items against non equality')
                return x >= value
        elif operator == '<=':
            def check(x):
                if not is_number(x):
                    return False;
                    # raise TypeError(
                    #     'excellib.countif() doesnt\'t work for checking non number items against non equality')
                return x <= value
        elif operator == '<>':
            def check(x):
                if not is_number(x):
                    return False;
                    # raise TypeError(
                    #     'excellib.countif() doesnt\'t work for checking non number items against non equality')
                return x != value
        else:
            def check(x):
                return x == criteria
    else:
        raise Exception('Could\'t parse criteria %s' % criteria)

    return check

def find_corresponding_index(rng, criteria):
    # parse criteria
    check = criteria_parser(criteria)
    valid = []
    for index, item in enumerate(rng):
        if check(item):
            valid.append(index)

    return valid

if __name__ == '__main__':
    pass



"""
Python equivalents of various excel functions
"""

# TODOMatthijs: waarom faalt de test hieronder?
# Foutmelding bij mij: ImportError: cannot import name 'Cell'

import math
import re
import numpy as np
from datetime import datetime
from math import log
from decimal import Decimal, ROUND_HALF_UP
from collections import Iterable, Counter
from converter.excel_util import flatten, is_number, is_collection, date_from_int, normalize_year, is_leap_year, \
    find_corresponding_index, is_nr, ERR_NA, ERR_DIV, rankdata, split_range, split_address, col2num, num2col, \
    evaluate_spreadsheet, to_values, is_valid, iterate_cells, get_range

# #####################################################################################
# A dictionary that maps excel function names onto python equivalents. You should
# only add an entry to this map if the python name is different to the excel name
# (which it may need to be to prevent conflicts with existing python functions
# with that name, e.g., max).

# So if excel defines a function foobar(), all you have to do is add a function
# called foobar to this module.  You only need to add it to the function map,
# if you want to use a different name in the python code. 

# Note: some functions (if, pi, atan2, and, or, array, ...) are already taken care of
# in the FunctionNode code, so adding them here will have no effect.
# Only add functions that have a different name in this file then the one in excel.
FUNCTION_MAP = {
    "ln": "xlog",
    "min": "xmin",
    "max": "xmax",
    "sum": "xsum",
    "gammaln": "lgamma",
    "round": "xround",
    "rank.eq": "rank_eq",
    "VerschilDekking": "verschildekking"
}

ERROR_VALUES = ['#DIV/0!', ERR_NA, '#NAME?', '#NULL!', '#NUM!', '#REF!', '#VALUE!']

######################################################################################
# List of excel equivalent functions
# TODO: needs unit testing

# TODOMatthijs: graag volgende test toevoegen:
# 8*None=0
# als je in excel iets met None vermenigvuld kopt er 0 uit, test maar, bij ons klapt hij.
# Dit ook graag testen voor: -, +, /, ... Kijk hoe excel daar mee om ga.
# We zullen dan bij een *, -, .. een functie er om heen moeten zetten die dan een 0 of andere waarde teruggeeft.
# Ik doe dat reeds met de safe_str functie hieronder, bij concatenation van string, zoek maar op "safe_str" hoe ik dat
# gebruik, ik denk daar ook die andere "safe" functies bij moeten maar misschien ergens anders.
# Ik heb ook voor *, /, +, - gemaakt (die gebruiken die safe_nr() hieronder)

def minus(val1, val2):
    return val1 - val2 if val1 and val2 else val1 if val1 else val2

def plus(val1, val2):
    return val1 + val2 if val1 and val2 else val1 if val1 else val2

def multiply(val1, val2):
    return val1 * val2 if val1 and val2 else 0

def divide(val1, val2):
    return val1 / val2 if val1 and val2 else 0 if not val1 else ERR_DIV

# def safe_nr(value):
#     return value if value else 0

def safe_str(value):
    return str(value) if value else ''

def row(address):
    col, row = [_f for _f in re.split('([A-Z $]+)', address) if _f]
    return int(row)

def substitute(spreadsheet, source, old, new, nth=None):
    if nth:
        find = source.find(old)
        # if find is not p1 we have found at least one match for the substring
        i = find != -1
        # loop util we find the nth or we find no match
        while find != -1 and i != nth:
            # find + 1 means we start at the last match start index + 1
            find = source.find(old, find + 1)
            i += 1
        # if i  is equal to nth we found nth matches so replace
        if i == nth:
            return source[:find] + new + source[find + len(old):]
        return source
    else:
        return source.replace(old, new)

def value(text):
    # make the distinction for naca numbers
    if text.find('.') > 0:
        return float(text)
    else:
        return int(text)

def xlog(a):
    if isinstance(a, (list, tuple, np.ndarray)):
        return [log(x) for x in flatten(a)]
    else:
        # print a
        return log(a)

# TODOMatthijs: maak een unit test voor deze functie en test met flex argumenten
def xmax(spreadsheet, *cell_address):
    if (not isinstance(cell_address, str)):
        items  = [x for x in cell_address if is_nr(x)]
        return max(items) if len(items) else ERR_NA # TODOMatthijs: ensure to test both these conditions
    else:
        result = None;
        def process(col_nr, row_nr, value):
            nonlocal result
            if (value is not None and is_nr(value) and (result is None or value > result)):
                result = value
            return True

        # Ensure to use the exact range
        rng = get_range(spreadsheet, cell_address)
        iterate_cells(spreadsheet, rng, process)
        # result = to_values(spreadsheet, rng, True, is_nr);
        return result
        # # ignore non numeric cells
        # data = [x for x in flatten(args) if isinstance(x, (int, float))]
        #
        # # however, if no non numeric cells, return zero (is what excel does)
        # if len(data) < 1:
        #     return 0
        # else:
        #     return max(data)

# TODOMatthijs: maak een unit test voor deze functie
def xmin(spreadsheet, *cell_address):
    if (not isinstance(cell_address, str)):
        items  = [x for x in cell_address if is_nr(x)]
        return min(items) if len(items) else ERR_NA # TODOMatthijs: ensure to test both these conditions
    else:
        result = None;
        def process(col_nr, row_nr, value):
            nonlocal result
            if (value is not None and is_nr(value) and (result is None or value < result)):
                result = value
            return True

        # Ensure to use the exact range
        rng = get_range(spreadsheet, cell_address)
        iterate_cells(spreadsheet, rng, process)
        # result = to_values(spreadsheet, rng, True, is_nr);
        return result

        # # ignore non numeric cells
        # data = [x for x in flatten(args) if isinstance(x, (int, float))]
        # # however, if no non numeric cells, return zero (is what excel does)
        # if len(data) < 1:
        #     return 0
        # else:
        # return min(data)

def xsum(spreadsheet, *args):
    result = 0;
    def process(col_nr, row_nr, value):
        nonlocal result
        if (value is not None and is_nr(value)):
            result += value
        return True

    if (is_nr(args)):
        return args

    # Ensure to use the exact range
    rng = None
    if is_collection(args):
        rng = []
        for ad in args:
            if is_nr(ad):
                result += ad
            else:
                rng.append(get_range(spreadsheet, ad))
    else:
        rng = args
    if is_collection(args):
        iterate_cells(spreadsheet, rng, process)
    # result = to_values(spreadsheet, rng, True, is_nr);
    return result

    # return sum(to_values(spreadsheet, rng, True, is_nr))

    # # ignore Non Numeric Cells
    # data = [x for x in flatten(args) if isinstance(x, (int))]
    # # however, if no non numeric cells, return zero (is what excel does)
    # if len(data) < 1:
    #     return 0
    # else:
    #     return sum(data)

"""
    Ref: https://support.office.com/en-us/article/SUMIF-function-169b8c99-c05c-4483-a712-1697a653039b
"""
def sumif(spreadsheet, cell_address, criteria, sum_range=None):
    rng = get_range(spreadsheet, cell_address)
    def is_valid_cell(value):
        return is_valid(value, criteria)
    return sum(to_values(spreadsheet, rng, True, is_valid_cell))

    # WARNING:
    # - wildcards not supported
    # - doesn't really follow 2nd remark about sum_range length

    # check = criteria_parser(criteria)
    # valid = []
    # for index, item in enumerate(rng):
    #     if check(item):
    #         valid.append(index)
    #
    # return valid

    # if sum_range is None:
    #     sum_range = []
    # if type(total_range) != list:
    #     raise TypeError('%s must be a list' % str(total_range))
    #
    # if type(sum_range) != list:
    #     raise TypeError('%s must be a list' % str(sum_range))
    #
    # if isinstance(criteria, list) and not isinstance(criteria, (str, bool)):  # ugly...
    #     return 0
    #
    # indexes = find_corresponding_index(total_range, criteria)
    #
    # def f(x):
    #     return sum_range[x] if x < len(sum_range) else 0
    #
    # if len(sum_range) == 0:
    #     return sum([total_range[x] for x in indexes])
    # else:
    #     return sum(map(f, indexes))

def average(spreadsheet, cell_address):
    result = 0;
    counter = 0
    def process(col_nr, row_nr, value):
        nonlocal result
        nonlocal  counter
        if (value is not None and is_nr(value)):
            result += value
            counter += 1
        return True

    # Ensure to use the exact range
    rng = get_range(spreadsheet, cell_address)
    iterate_cells(spreadsheet, rng, process)
    return result / counter

    # sheet, start, end = split_range(rng)
    # sh, start_col, start_row = split_address(start)
    # sh, end_col, end_row = split_address(end)
    # start_col_nr, end_col_nr = col2num(start_col), col2num(end_col)
    # start_row_nr, end_row_nr = int(start_row), int(end_row)
    # result = 0
    # for row_nr in range(start_row_nr, end_row_nr + 1):
    #     for col_nr in range(start_col_nr, end_col_nr + 1):
    #         value = evaluate_spreadsheet(spreadsheet, sheet, num2col(col_nr), row_nr)
    #         if (isinstance(value, (int, float))):
    #             result += value
    # return result
    # l = list(flatten(*args))

# TODOMatthijs: make a unit test please, ensure it deals with flexible argument (see sum function how it's done)
def right(text, n):
    if isinstance(text, str) or isinstance(text, str):
        return text[-n:]
    else:
        # TODO: get rid of the decimal
        return str(int(text))[-n:]

"""
Ref: https://support.office.com/en-us/article/INDEX-function-a5dcf0dd-996d-40a4-a822-b56b061328bd
"""
def index(spreadsheet, cell_address, row_nr=None, col_nr=None):
    rng = get_range(spreadsheet, cell_address)
    sheet, start, end = split_range(rng)
    sh, start_col, start_row = split_address(start)
    if (row_nr and col_nr):
        col = num2col(col2num(start_col) + col_nr - 1)
        row = int(start_row) + row_nr - 1
    elif row_nr:
        sh, end_col, end_row = split_address(end)
        if start_col == end_col:  # it concerns a vertical array
            col = start_col
            row = int(start_row) + row_nr - 1
        elif start_row == end_row:  # it concerns a horizontal array
            col = num2col(col2num(start_col) + row_nr - 1)
            row = start_row
        else:
            raise Exception(
                "Multi dimension array not allowed with only row or column, cell_address: [%s], col_nr: [%s], row_nr: [%s]" % (
                    cell_address, col_nr))
    else:
        raise Exception("Both row_nr and col_nr are missing, cell_address: [%s]" % (cell_address))
    return evaluate_spreadsheet(spreadsheet, sheet, col, row)

    # array = args[0]
    # row = args[1]
    #
    # if len(args) == 3:
    #     col = args[2]
    # else:
    #     col = 1
    #
    # if isinstance(array[0], (list, tuple, np.ndarray)):
    #     # rectangular array
    #     array[row - 1][col - 1]
    # elif row == 1 or col == 1:
    #     return array[row - 1] if col == 1 else array[col - 1]
    # else:
    #     raise Exception("index (%s,%s) out of range for %s" % (row, col, array))

"""
Ref: https://support.office.com/en-us/article/RANK-EQ-function-284858ce-8ef6-450e-b662-26245be04a40
"""
def rank_eq(spreadsheet, item, cell_address, mode):
    rng = get_range(spreadsheet, cell_address)
    sheet, start, end = split_range(rng)
    sh, start_col, start_row = split_address(start)
    sh, end_col, end_row = split_address(end)
    start_col_nr, start_row_nr = col2num(start_col), int(start_row)
    end_row_nr = int(end_row)

    if start_col != end_col:
        raise Exception("The start and end column are expected to be the same")

    items = []
    for row_nr in range(start_row_nr, end_row_nr + 1):
        items.append(evaluate_spreadsheet(spreadsheet, sheet, num2col(start_col_nr), row_nr))

    # items = [x for x in items if x is not None]
    if (mode == 1):
        return rankdata(items, 'min')[items.index(item)]
    else:
        raise Exception("Mode isn't supported yet, mode: [%s], item: [%s]" % (mode, item))

# TODOMatthijs: make a unit test please
def lookup(value, lookup_range, result_range):
    if not isinstance(value, (int, float)):
        raise Exception("Non numeric lookups (%s) not supported" % value)

    # TODO: note, may return the last equal value

    # index of the last numeric value
    lastnum = -1
    for i, v in enumerate(lookup_range):
        if isinstance(v, (int, float)):
            if v > value:
                break
            else:
                lastnum = i

    if lastnum < 0:
        raise Exception("No numeric data found in the lookup range")
    else:
        if i == 0:
            raise Exception("All values in the lookup range are bigger than %s" % value)
        else:
            if i >= len(lookup_range) - 1:
                # return the biggest number smaller than value
                return result_range[lastnum]
            else:
                return result_range[i - 1]

"""
It will convert the list of rows to a dictionary such that the keys correspond 
to the first column in the list of row and the values to the column indicated through the 
argument col index.
"""
def to_dictionary(row_list, col_index_num):
    return {row[0]: row[col_index_num] for row in row_list}

# def hlookup_row(item, row_dict, case_sensitive=False):
#     if not case_sensitive:
#         for key in row_dict.keys():
#             if str(key).lower() == str(item).lower():
#                 return row_dict[key]
#     else:
#         return row_dict[item]

"""
The VLOOKUP function performs a vertical lookup by searching for a value in the first column of a table and 
returning the value in the same row in the index_number position.
Ref: https://www.techonthenet.com/excel/formulas/vlookup.php
"""
# TODOMatthijs: unit test dit goed: ook die approximate setting (see ref)
def vlookup(spreadsheet, item, cell_address, col_index_nr, approximate_match=False):
    if not item:
        return ERR_NA

    if (approximate_match and not isinstance(item, (float, int))):
        raise Exception("Item must be a number: [%s], row_index_nr: [%s]" % (item, col_index_nr))

    # Ensure to use the exact range
    rng = get_range(spreadsheet, cell_address)
    item = str(item).lower()
    sheet, start, end = split_range(rng)
    sh, start_col, start_row = split_address(start)
    sh, end_col, end_row = split_address(end)
    search_index = num2col(col2num(start_col) + col_index_nr - 1)
    previous = None
    ln = int(end_row) - int(start_row) + 1
    for row_nr in range(int(start_row), int(end_row) + 1):
        value = evaluate_spreadsheet(spreadsheet, sheet, start_col, row_nr)
        if (value is None):  # we skip empty cells/columns
            continue
        value = str(value).lower()
        if approximate_match and previous and previous > value:
            raise Exception(
                "Items not in ascending order. previous: [%s], current: [%s], item: [%s], row_index_r: [%s]"
                % (previous, value, item, col_index_nr))

        if value == item or approximate_match and (
                                ln - 1 == row_nr and value < item or previous and value > item and previous < item):
            return evaluate_spreadsheet(spreadsheet, sheet, search_index, row_nr)

        previous = value
    return ERR_NA
    # col_values = to_dictionary(rng, col_index_num - 1)
    # return vlookup_column(item, col_values, approximate_match)

"""
":" means Start:End
Finds the specified col_index_num in excel rows and selects that row
and creates a dictionary of the specified row
"""
def create_column_dict(excel_rows, col_index_num, case_sensitive=False):
    if case_sensitive:
        return {row: row[col_index_num] for row in excel_rows}
    else:
        return {row[0]: row[col_index_num] for row in excel_rows}

"""
The Excel HLOOKUP function performs a horizontal lookup by searching for a value in the top row of the table and 
returning the value in the same column based on the index_number.
Ref: https://www.techonthenet.com/excel/formulas/hlookup.php
"""
# TODOMatthijs: unit test dit goed: ook die approximate setting (see ref), Voeg een unit test waarom verschil in kleine/letters
# wordt gebruikt, d.w.z: zoeken op grote letters terwijl de waarde in de range hetzelfde is maar in grote letters bijvoorbeeld.
# Test ook hlookup en vlookup met if_error a.u.b.
def hlookup(spreadsheet, item, cell_address, row_index_nr, approximate_match=False):
    if not item:
        return ERR_NA

    # rows = [x for x in rows]
    if (approximate_match and not isinstance(item, (float, int))):
        raise Exception("Item must be a number: [%s], row_index_nr: [%s]" % (item, row_index_nr))

    # Ensure to use the exact range
    rng = get_range(spreadsheet, cell_address)
    item = str(item).lower()
    sheet, start, end = split_range(rng)
    sh, start_col, start_row = split_address(start)
    sh, end_col, end_row = split_address(end)
    start_col_nr, end_col_nr = col2num(start_col), col2num(end_col)
    ln = end_col_nr - start_col_nr + 1
    previous = None
    for col_nr in range(start_col_nr, end_col_nr + 1):
        value = evaluate_spreadsheet(spreadsheet, sheet, num2col(col_nr), start_row)
        if (value is None):  # we skip empty cells/columns
            continue

        value = str(value).lower()
        if approximate_match and previous and int(previous) > int(value):
            raise Exception(
                "Items not in ascending order. previous: [%s], current: [%s], item: [%s], row_index_r: [%s]"
                % (previous, value, item, row_index_nr))

        if value == item or approximate_match and (
                                ln - 1 == col_nr and value < item or previous and value > item and previous < item):
            return evaluate_spreadsheet(spreadsheet, sheet, num2col(col_nr), int(start_row) + row_index_nr - 1)

        previous = value
    return ERR_NA

    # if (str(rows[0][i]).lower() == str(item).lower() if approximate_match else str(rows[0][i]) == str(item)):
    #     return rows[row_index_num - 1][i]
    # return None

# Change item from Uppercase to Lowercase
# TODOEd: shouldn't this be moves to excel_util?
def lower_case_item(item):
    if isinstance(item, Iterable):
        item = [str(entry).lower() for entry in item]
        return item
    else:
        return str(item).lower()

# TODOMatthijs: make a unit test please
def linest(*args, **kwargs):
    y = args[0]
    x = args[1]

    if len(args) == 3:
        const = args[2]
        if isinstance(const, str):
            const = (const.lower() == "true")
    else:
        const = True

    degree = kwargs.get('degree', 1)

    # build the vandermonde matrix
    a = np.vander(x, degree + 1)

    if not const:
        # force the intercept to zero
        a[:, -1] = np.zeros((1, len(x)))

    # perform the fit
    (coefs, residuals, rank, sing_vals) = np.linalg.lstsq(a, y)

    return coefs

# TODOMatthijs: make a unit test please
def npv(*args):
    discount_rate = args[0]
    cashflow = args[1]
    return sum([float(x) * (1 + discount_rate) ** -(i + 1) for (i, x) in enumerate(cashflow)])

def match_type_convert(value):
    if type(value) == str:
        value = value.lower()
    elif type(value) == int:
        value = float(value)
    return value

# TODOMatthijs: maak a.u.b. een goede matsch test a.u.b.: ook met boundary testen (zie de excel_lib_test)
# Voorbeeld van boundary testen: dat match_type=1, en de gezochte waarde is exact gelijk en bevindt zich in de laatste rij, of
# juist niet exact, maar toch de juiste (zie ref)
# Ref: https://www.techonthenet.com/excel/formulas/match.php
#
# TODOMatthijs: toevoegen van een test waarbij een match voor horizontale kolom wordt gebruikt.
# TODOEd: add support for: match_type=1 and value exact not found, but correct return value present in the last row.
def match(spreadsheet, lookup_value, cell_address, match_type=1):
    previous = None
    result = None
    lookup_value = match_type_convert(lookup_value)

    def process(col_nr, row_nr, value):
        nonlocal previous
        nonlocal result
        nonlocal match_type
        nonlocal use_col
        if (match_type == 1):
            value = match_type_convert(value);
            if (previous and value < previous):
                raise Exception(
                    "Array must be in ascending order for match type 1, previous: [%s], next: [%s], lookup value: [%s], range: [%s]" % (
                        previous, next, lookup_value, rng))
            if (lookup_value == value):
                result = col_nr if use_col else row_nr
                return False
            elif (lookup_value < value and previous and lookup_value > previous):
                result -= 1
                return False
        elif (match_type == 0):
            if (lookup_value == match_type_convert(value)):
                result = col_nr if use_col else row_nr
                return False
        else:
            value = match_type_convert(value);
            if (previous and value > previous):
                raise Exception(
                    "Array must be in descending order for match type -1, previous: [%s], next: [%s], lookup value: [%s], range: [%s]" % (
                        previous, next, lookup_value, rng))
            if (lookup_value == value):
                result = col_nr if use_col else row_nr
                return False
            elif (lookup_value > value and previous and lookup_value < previous):
                result -= 1
                return False

        previous = value
        return True

    rng = get_range(spreadsheet, cell_address)
    sheet, start, end = split_range(rng)
    sh, start_col, start_row = split_address(start)
    sh, end_col, end_row = split_address(end)
    if start_col == end_col:
        use_col = False
    elif start_row == end_row:
        use_col = True
    else:
        raise Exception("Match only supports a single search row or column. Address: [%s]"(rng))
    iterate_cells(spreadsheet, rng, process)
    if result:
        return result - col2num(start_col) + 1 if use_col else result - int(start_row) + 1
    return ERR_NA;

    # if match_type == 1:
    #     # Verify ascending sort
    #     posmax = -1
    #     for i in range((len(lookup_array))):
    #         current = type_convert(lookup_array[i])
    #
    #         if i is not len(lookup_array) - 1 and current > type_convert(lookup_array[i + 1]):
    #             raise Exception('for match_type 0, lookup_array must be sorted ascending')
    #         if current <= lookup_value:
    #             posmax = i
    #     if posmax == -1:
    #         return ERR_NA
    #     # raise Exception('no result in lookup_array for match_type 0')
    #     return posmax + 1  # Excel starts at 1
    #
    # elif match_type == 0:
    #     # No string wildcard
    #     search = [type_convert(x) for x in lookup_array]
    #     if lookup_value in search:
    #         return search.index(lookup_value) + 1
    #     return ERR_NA
    #
    # # return [type_convert(x) for x in lookup_array].index(lookup_value) + 1
    #
    # elif match_type == -1:
    #     # Verify descending sort
    #     posmin = -1
    #     for i in range((len(lookup_array))):
    #         current = type_convert(lookup_array[i])
    #
    #         if i is not len(lookup_array) - 1 and current < type_convert(lookup_array[i + 1]):
    #             raise Exception("for match_type 0, lookup_array must be sorted descending")
    #         if current >= lookup_value:
    #             posmin = i
    #     if posmin == -1:
    #         # TODOMatthijs: zie opm hierboven
    #         return ERR_NA
    #     # raise Exception('no result in lookup_array for match_type 0')
    #     return posmin + 1  # Excel starts at 1

# TODOMatthijs: make a unit test please
# Excel Reference: https://support.office.com/en-us/article/MOD-function-9b6cd169-b6ee-406a-a97b-edf2a9dc24f3
def mod(nb, q):
    if not isinstance(nb, int):
        raise TypeError("%s is not an integer" % str(nb))
    elif not isinstance(q, int):
        raise TypeError("%s is not an integer" % str(q))
    else:
        return nb % q

# TODOMatthijs: make a unit test please (voor verschillende omstandigheden)
# Excel reference: https://support.office.com/en-us/article/COUNT-function-a59cd7fc-b623-4d93-87a4-d23bf411294c
def count(spreadsheet, cell_address):
    counter = 0;
    def process(col_nr, row_nr, value):
        nonlocal counter
        if (is_nr(value)):
            counter += 1
        return True

    # Ensure to use the exact range
    rng = get_range(spreadsheet, cell_address)
    iterate_cells(spreadsheet, rng, process)
    # result = to_values(spreadsheet, rng, True, is_nr);
    return counter

    # l = list(args)
    # total = 0
    # for arg in l:
    #     if type(arg) == list:
    #         total += len([x for x in arg if is_number(x) and type(x) is not bool])  # count inside a list
    #     elif is_number(arg):  # int() is used for text representation of numbers
    #         total += 1
    #
    # return total

# TODOMatthijs: unit tests van onderstaande toevoegen.
def iferror(spreadsheet, value, value_if_error):
    # list of errors that are used in excel
    return value_if_error if value in ERROR_VALUES else value

"""
Counta function: counts the non empty cells
Ref: https://exceljet.net/excel-functions/excel-counta-function
"""
def counta(spreadsheet, cell_address):
    counter = 0;
    def process(col_nr, row_nr, value):
        nonlocal counter
        if (value is not None):
            counter += 1
        return True

    # Ensure to use the exact range
    rng = get_range(spreadsheet, cell_address)
    iterate_cells(spreadsheet, rng, process)
    # result = to_values(spreadsheet, rng, True, is_nr);
    return counter

    # return len([x for x in rng if x is not None])

# TODOMathij: a.u.b. de unit test verder uitbreiden (ik heb al een beginnetje gemaakt)
# Excel reference: https://support.office.com/en-us/article/COUNTIF-function-e0de10c6-f885-4e71-abb4-1f464816df34
def countif(spreadsheet, cell_address, criteria):
    # WARNING:
    # - wildcards not supported
    # - support of strings with >, <, <=, =>, <> not provided
    rng = get_range(spreadsheet, cell_address)
    def is_valid_cell(value):
        return is_valid(value, criteria)
    return len(to_values(spreadsheet, rng, True, is_valid_cell))

    # valid = find_corresponding_index(range, criteria)
    # return len(valid)

# def find_corresponding_index(rng, criteria):
#     # parse criteria
#     check = criteria_parser(criteria)
#     valid = []
#     for index, item in enumerate(rng):
#         if check(item):
#             valid.append(index)
#
#     return valid


# TODOMathijs: please add a unit test
# Ref: https://support.office.com/en-us/article/COUNTIFS-function-dda3dc6e-f74e-4aee-88bc-aa8c2a866842
def countifs(*args):
    arg_list = list(args)
    l = len(arg_list)

    if l % 2 != 0:
        raise Exception('excellib.countifs() must have a pair number of arguments, here %d' % l)

    if l >= 2:
        indexes = find_corresponding_index(args[0], args[1])  # find indexes that match first layer of countif

        remaining_ranges = [elem for i, elem in enumerate(arg_list[2:]) if i % 2 == 0]  # get only ranges
        remaining_criteria = [elem for i, elem in enumerate(arg_list[2:]) if i % 2 == 1]  # get only criteria

        filtered_remaining_ranges = []

        for range in remaining_ranges:  # filter items in remaining_ranges that match valid indexes from first countif layer
            filtered_remaining_range = []

            for index, item in enumerate(range):
                if index in indexes:
                    filtered_remaining_range.append(item)

            filtered_remaining_ranges.append(filtered_remaining_range)

        new_tuple = ()

        for index, range in enumerate(
                filtered_remaining_ranges):  # rebuild the tuple that will be the argument of next layer
            new_tuple += (range, remaining_criteria[index])

        return min(countifs(*new_tuple), len(indexes))  # only consider the minimum number across all layer responses

    else:
        return float('inf')

# Ref: https://support.office.com/en-us/article/ROUND-function-c018c5d8-40fb-4053-90b1-b3e7f61a213c
def xround(spreadsheet, number, num_digits=0):
    if not is_number(number):
        raise TypeError("%s is not a number" % str(number))
    if not is_number(num_digits):
        raise TypeError("%s is not a number" % str(num_digits))

    if num_digits >= 0:  # round to the right side of the point
        return float(Decimal(repr(number)).quantize(Decimal(repr(pow(10, -num_digits))), rounding=ROUND_HALF_UP))
        # see https://docs.python.org/2/library/functions.html#round
        # and https://gist.github.com/ejamesc/cedc886c5f36e2d075c5

    else:
        return round(number, num_digits)

# TODOMatthijs: kunnen we dit niet direct op math.ceil van python mappen (op andere plek in de code)
def roundup(spreadsheet, number, num_digits=0):
    if num_digits != 0:
        raise Exception(
            "Currently only rounding up with 0 digits is supported, number: [%s], digits: [%s]" % (number, num_digits))
    return math.ceil(number)

def rounddown(spreadsheet, number, num_digits=0):
    if num_digits != 0:
        raise Exception(
            "Currently only rounding down with 0 digits is supported, number: [%s], digits: [%s]" % (
                number, num_digits))
    return math.trunc(number)

# TODOMathijs: please add a unit test
# Excel reference: https://support.office.com/en-us/article/MID-MIDB-functions-d5f9e25c-d7d6-472e-b568-4ecb12433028
def mid(text, start_num, num_chars):
    text = str(text)
    if type(start_num) != int:
        raise TypeError("%s is not an integer" % str(start_num))
    if type(num_chars) != int:
        raise TypeError("%s is not an integer" % str(num_chars))

    if start_num < 1:
        raise ValueError("%s is < 1" % str(start_num))
    if num_chars < 0:
        raise ValueError("%s is < 0" % str(num_chars))

    return text[start_num:num_chars]

# Ref: https://support.office.com/en-us/article/DATE-function-e36c0c8c-4104-49da-ab83-82328b832349
def date(year, month, day):
    if type(year) != int:
        raise TypeError("%s is not an integer" % str(year))

    if type(month) != int:
        raise TypeError("%s is not an integer" % str(month))

    if type(day) != int:
        raise TypeError("%s is not an integer" % str(day))

    if year < 0 or year > 9999:
        raise ValueError("Year must be between 1 and 9999, instead %s" % str(year))

    if year < 1900:
        year = 1900 + year

    year, month, day = normalize_year(year, month, day)  # taking into account negative month and day values

    date_0 = datetime(1900, 1, 1)
    date = datetime(year, month, day)

    result = (datetime(year, month, day) - date_0).days + 2

    if result <= 0:
        raise ArithmeticError("Date result is negative")
    else:
        return result

# TODOMathijs: please add a unit test
# Excel reference: https://support.office.com/en-us/article/YEARFRAC-function-3844141e-c76d-4143-82b6-208454ddc6a8
def yearfrac(start_date, end_date, basis=0):
    def actual_nb_days_ISDA(start, end):  # needed to separate days_in_leap_year from days_not_leap_year
        y1, m1, d1 = start
        y2, m2, d2 = end

        days_in_leap_year = 0
        days_not_in_leap_year = 0

        year_range = list(range(y1, y2 + 1))

        for y in year_range:

            if y == y1 and y == y2:
                nb_days = date(y2, m2, d2) - date(y1, m1, d1)
            elif y == y1:
                nb_days = date(y1 + 1, 1, 1) - date(y1, m1, d1)
            elif y == y2:
                nb_days = date(y2, m2, d2) - date(y2, 1, 1)
            else:
                nb_days = 366 if is_leap_year(y) else 365

            if is_leap_year(y):
                days_in_leap_year += nb_days
            else:
                days_not_in_leap_year += nb_days

        return days_not_in_leap_year, days_in_leap_year

    def actual_nb_days_AFB_alter(start,
                                 end):  # http://svn.finmath.net/finmath%20lib/trunk/src/main/java/net/finmath/time/daycount/DayCountConvention_ACT_ACT_YEARFRAC.java
        y1, m1, d1 = start
        y2, m2, d2 = end

        delta = date(*end) - date(*start)

        if delta <= 365:
            if is_leap_year(y1) and is_leap_year(y2):
                denom = 366
            elif is_leap_year(y1) and date(y1, m1, d1) <= date(y1, 2, 29):
                denom = 366
            elif is_leap_year(y2) and date(y2, m2, d2) >= date(y2, 2, 29):
                denom = 366
            else:
                denom = 365
        else:
            year_range = list(range(y1, y2 + 1))
            nb = 0

            for y in year_range:
                nb += 366 if is_leap_year(y) else 365

            denom = nb / len(year_range)

        return delta / denom

    if not is_number(start_date):
        raise TypeError("start_date %s must be a number" % str(start_date))
    if not is_number(end_date):
        raise TypeError("end_date %s must be number" % str(end_date))
    if start_date < 0:
        raise ValueError("start_date %s must be positive" % str(start_date))
    if end_date < 0:
        raise ValueError("end_date %s must be positive" % str(end_date))

    if start_date > end_date:  # switch dates if start_date > end_date
        temp = end_date
        end_date = start_date
        start_date = temp

    y1, m1, d1 = date_from_int(start_date)
    y2, m2, d2 = date_from_int(end_date)

    if basis == 0:  # US 30/360
        d2 = 30 if d2 == 31 and (d1 == 31 or d1 == 30) else min(d2, 31)
        d1 = 30 if d1 == 31 else d1

        count = 360 * (y2 - y1) + 30 * (m2 - m1) + (d2 - d1)
        result = count / 360

    elif basis == 1:  # Actual/actual
        result = actual_nb_days_AFB_alter((y1, m1, d1), (y2, m2, d2))

    elif basis == 2:  # Actual/360
        result = (end_date - start_date) / 360

    elif basis == 3:  # Actual/365
        result = (end_date - start_date) / 365

    elif basis == 4:  # Eurobond 30/360
        d2 = 30 if d2 == 31 else d2
        d1 = 30 if d1 == 31 else d1

        count = 360 * (y2 - y1) + 30 * (m2 - m1) + (d2 - d1)
        result = count / 360

    else:
        raise ValueError("%d must be 0, 1, 2, 3 or 4" % basis)

    return result

# TODO: This function might need more solid testing
def isNa(value):
    try:
        eval(value)
        return False
    except:
        return True

def verschildekking(minvalue, rijvalue):
    lager = 0
    hoger = 0
    verschil = 0
    data = ""

    if rijvalue or minvalue is not None:
        if rijvalue < minvalue:
            lager += 1
            verschil = np.array(rijvalue) - np.array(minvalue)
        if rijvalue > minvalue:
            hoger += 1
    if lager == 0 and hoger == 0:
        data += "Vergelijkbaar"
    elif lager == 1 and hoger is None and verschil == -1:
        data += "Slechter1"
    elif lager > 1 and hoger is None:
        data += "Slechter"
    elif lager == 1 and hoger > 0:
        data += "Anders"
    elif lager == 0 and hoger > 0:
        data += "Beter"

    return data

def PositieMinimumAls(criteriumbereik, criterium, minimum, minimumbereik):
    minpositie = 0
    while criteriumbereik is not None:
        if criteriumbereik == criterium:
            if minimum or minimumbereik < minimum:
                minimum = minimumbereik
            minpositie = i
        i = i + 1
    return minpositie

def PositieBucketMinder(criteriumbereik, bucketbereik, prijsbereik):
    i = 1
    minpositie = 0

    while criteriumbereik(i) is not None:
        if criteriumbereik(i).value == "Anders1" or criteriumbereik(i).value == "Slechter1":
            if bucketbereik(i) - prijsbereik(i) < minimum:
                if minimum is not None or prijsbereik(i) < minimum:
                    minimum = prijsbereik(i)
                minpositie = i
                i = i + 1
    return minpositie

def PositieMinBucketIncVervang(huidigebuckets, vervangwaarde, vervangpositie, allebuckets, prijsrange):
    minpositie = 0
    minprijs = 9999
    looprij = 1
    eindrij = allebuckets
    aantalbuckets = huidigebuckets

    while looprij < eindrij:
        if allebuckets(looprij, vervangpositie).value == vervangwaarde:
            bucket = 1
            geldig = 1
            while bucket <= aantalbuckets:
                if bucket != vervangpositie:
                    if allebuckets(looprij, bucket).value < huidigebuckets(1, bucket).value:
                        geldig = 0
                        bucket = aantalbuckets
                    bucket = bucket + 1

                if geldig == 1 and prijsrange(looprij, 1).value < minprijs:
                    minprijs = prijsrange(looprij, 1).value
                    minpositie = looprij

                looprij = looprij + 1
        return minpositie

if __name__ == '__main__':
    pass



# Excel compiler.
# creates nodes from cells puts them in a graph and then puts that graph in a cellmap.
# Has several functions to set and calculate the changed values

import logging
import networkx as nx
from networkx.classes.digraph import DiGraph
from .excel_node import *
from .excel_util import *
from .excel_util import Cell
from .excel_lib import *  # Used dynamically in compiler
from .excel_wrapper import ExcelOpxWrapper as ExcelWrapperImpl
from .tokenizer import ExcelParser, f_token

# Set the sheet into a node
def add_node_to_graph(g, n):
    g.add_node(n)
    g.node[n]['sheet'] = n.sheet

    if isinstance(n, Cell):
        g.node[n]['label'] = n.col + str(n.row)
    else:
        # strip the sheet
        g.node[n]['label'] = n.address()[n.address().find('!') + 1:]

class ExcelCompiler(object):
    def __init__(self, filename=None, *args, **kwargs):
        # Load file
        super(ExcelCompiler, self).__init__()
        self.filename = filename
        # Set the filename
        self.excel = ExcelWrapperImpl(filename=filename)
        self.excel.connect()

        self.log = logging.getLogger("decode.{0}".format(self.__class__.__name__))

    def cell2code(self, cell):
        # Generate python code for the given cell
        if cell.formula:
            e = shunting_yard(cell.formula or str(cell.value))
            if e is None or len(e) == 0:
                raise Exception("Incorrect expression: [%s], cell: [%s], cell.formula: [%s], cell.value: [%s]"
                                % (e, cell, cell.formula, cell.value))

            ast, root = build_ast(e)
            code = root.emit(ast, context=Context(cell, self.excel))
        elif isinstance(cell.value, str) and re.match('[A-Za-z]+', cell.value):
            ast = None
            code = cell.value
        else:
            ast = None
            code = str('"' + cell.value + '"' if isinstance(cell.value, str) else cell.value)

        return code, ast

    def gen_graph(self, range, sheet=None):
        return self.gen_graph_multi({sheet: range})

    """
    Example usage:
        gen_graph({'Output1': 'A1:B5', 'Output2': 'A1:E8'})
        gen_graph({'Output1': '[A1:B5, AB4:FE85]', 'Output2': 'A1:E8'})
    """
    def gen_graph_multi(self, ranges):
        # When given a starting point (e.g., A6 or A3:B7) on a particular sheet,
        # generate a Spreadsheet instance that captures the logic and control flow of the equations

        # starting points
        # cursheet = sheet if sheet else self.excel.get_active_sheet()
        # self.excel.set_sheet(cursheet)

        # no need to output NumberRows and NumberColoms here, since seed can be a list of unlinked cells
        seeds = None
        try:
            seeds = Cell.make_cells_multi(self.excel, ranges)
            print(("Ranges [%s] expanded into [%s] cell_len" % (ranges, len(seeds))))
            return self.gen_graph_intern(seeds)
        except Exception as ex:
            raise Exception("Error making cells, ranges: [%s], excel: [%s]" % (ranges, self.excel)) from ex

    def gen_graph_intern(self, seeds):
        # only keep seeds with formulas or numbers
        # Scan seeds for data_types int, float or strings
        seeds = [s for s in seeds if s.formula or isinstance(s.value, (int, float, str))]

        print(("Filtered seeds (keeping formulas, strings, ints, floats): [%s] " % len(seeds)))

        # cells to analyze: only formulas or strings
        # We added the string to the TODOlist but it doesn't seem to be needed afterwards, such that it can be removed.
        # todo = [s for s in seeds if s.formula or isinstance(s.value, str)]
        todo = [s for s in seeds if s.formula]

        print(("Todo cells: [%s]" % len(todo)))

        # map of all cells
        cellmap = dict([(x.address(), x) for x in seeds])

        # directed graph
        g = nx.DiGraph()

        # match the info in cellmap
        for c in list(cellmap.values()):
            add_node_to_graph(g, c)

        cursheet = self.excel.get_active_sheet()
        while todo:
            c1 = todo.pop()

            # set the current sheet so relative addresses resolve properly
            if c1.sheet != cursheet:
                cursheet = c1.sheet
                self.excel.set_sheet(cursheet)

            try:
                pystr, ast = self.cell2code(c1)  # parse the formula into code
                c1.python_expression = pystr  # set the code & compile it (will flag problems sooner rather than later)
                c1.compile()
            except Exception as ex:
                raise Exception(
                    "Error compiling/cell2code cell: [%s], todo: [%s]" % (c1, todo)) from ex

            if ast is not None:
                # get all the cells/ranges this formula refers to
                deps = [x.tvalue.replace('$', '') for x in ast.nodes() if isinstance(x, RangeNode)]
                deps = uniqueify(deps)  # remove dupes

                for dep in deps:
                    # Ensure to always set the active sheet, especially in case no dependency sheet is found, that means
                    # the dependency occurs on the same sheet as cursheet, but in case we just had a dependency to another
                    # sheet that can result in using the wrong sheet if not set.
                    dep_sheet = create_sheet(dep)
                    dep_sheet = dep_sheet if dep_sheet else cursheet
                    self.excel.set_sheet(dep_sheet if dep_sheet else cursheet)

                    # if the dependency is a multi-cell range, create a range object
                    if is_range(dep):
                        # this will make sure we always have an absolute address
                        rng_address = create_address(dep, sheet=dep_sheet)
                        if rng_address in cellmap:
                            # already dealt with this range
                            # add an edge from the range to the parent
                            g.add_edge(cellmap[rng_address], cellmap[c1.address()])
                            continue
                        else:
                            # turn into cell objects
                            rng = None
                            try:
                                # TODOEd: should we still create the cells and rng_value and store in the range?? as we supply
                                # the range as string to the internal functions (see Redmine issues)
                                cells, nrows, ncols, rng_details = Cell.make_cells(self.excel, dep, sheet=dep_sheet)
                                # get the values so we can set the range value
                                rng_value = None
                                if nrows == 1 or ncols == 1:
                                    rng_value = [c.value for c in cells]
                                else:
                                    rng_value = [[c.value for c in cells[i]] for i in range(len(cells))]

                                # update the original range as the returned one is probably mor specific if it wasn't a list
                                rng = CellRange(rng_details if rng_details else rng_address, rng_value, sheet=dep_sheet)
                                # cells, nrows, ncols = Cell.make_cells(self.excel, {cursheet: dep})
                            except Exception as ex:
                                raise Exception(
                                    "Error making cells, cell: [%s], ast: [%s], dep: [%s], rng address: [%s], rng: [%s], todo_len: [%s], seeds_len: [%s]" %
                                    (c1, ast, dep, rng_address, rng, len(todo), len(seeds))) from ex

                            # save the range
                            # Ensure to use the original address to store it as other need it to find it.
                            cellmap[rng_address] = rng
                            # cellmap[rng.address()] = rng
                            # add an edge from the range to the parent
                            add_node_to_graph(g, rng)
                            g.add_edge(rng, cellmap[c1.address()])
                            # cells in the range should point to the range as their parent
                            target = rng
                    else:
                        # not a range, create the cell object
                        cells = (Cell.resolve_cell(self.excel, dep, sheet=dep_sheet),)
                        target = cellmap[c1.address()]

                    # process each cell
                    for c2 in flatten(cells):
                        # Check that cells are None
                        if c2 is not None:
                            # if we havent treated this cell already
                            if c2.address() not in cellmap:
                                if c2.formula:
                                    # cell with a formula, needs to be added to the todo list
                                    todo.append(c2)
                                    # print("appended ", c2.address())

                                else:
                                    try:
                                        # constant cell, no need for further processing, just remember to set the code
                                        pystr, ast = self.cell2code(c2)
                                        c2.python_expression = pystr
                                        c2.compile()
                                    except Exception as ex:
                                        raise Exception(
                                            "Error compiling/cell2code dependency cell, cell dep: [%s] cell: [%s], ast: [%s], dep: [%s], "
                                            "todo: [%s], seeds: [%s]" % (c2, c1, ast, dep, todo, seeds)) from ex
                                # print("skipped ", c2.address())

                                # save in the cellmap
                                cellmap[c2.address()] = c2
                                # add to the graph
                                add_node_to_graph(g, c2)

                            # add an edge from the cell to the parent (range or cell)
                            g.add_edge(cellmap[c2.address()], target)

        print(("Graph construction done, %s nodes, %s edges, %s cellmap entries" % (
            len(g.nodes()), len(g.edges()), len(cellmap))))
        sp = Spreadsheet(g, cellmap)
        return sp

def create_address(rng, sheet=None):
    address = rng.replace('$', '')
    if sheet is not None and rng.find('!') < 0:
        address = sheet + "!" + address
    return address

def is_continue_Required(self, cell):
    cell is not None and cell.formula is None

def shunting_yard(expression):
    """
    Tokenize an excel formula expression into reverse polish notation

    Core algorithm taken from wikipedia with varargs extensions from
    http://www.kallisti.net.nz/blog/2008/02/extension-to-the-shunting-yard-algorithm-to-allow-variable-numbers-of-arguments-to-functions/
    """
    # remove leading =

    if expression.startswith('='):
        expression = expression[1:]

    p = ExcelParser()
    p.parse(expression)

    # insert tokens for '(' and ')', to make things clearer below
    tokens = []
    for t in p.tokens.items:
        if t.ttype == "function" and t.tsubtype == "start":
            t.tsubtype = ""
            tokens.append(t)
            tokens.append(f_token('(', 'arglist', 'start'))
        elif t.ttype == "function" and t.tsubtype == "stop":
            tokens.append(f_token(')', 'arglist', 'stop'))
        elif t.ttype == "subexpression" and t.tsubtype == "start":
            t.tvalue = '('
            tokens.append(t)
        elif t.ttype == "subexpression" and t.tsubtype == "stop":
            t.tvalue = ')'
            tokens.append(t)
        else:
            tokens.append(t)

    # print "tokens: ", "|".join([x.tvalue for x in tokens])

    # http://office.microsoft.com/en-us/excel-help/calculation-operators-and-precedence-HP010078886.aspx
    operators = {':': Operator(':', 8, 'left'), '': Operator(' ', 8, 'left'), ',': Operator(',', 8, 'left'),
                 'u-': Operator('u-', 7, 'left'), '%': Operator('%', 6, 'left'), '^': Operator('^', 5, 'left'),
                 '*': Operator('*', 4, 'left'), '/': Operator('/', 4, 'left'), '+': Operator('+', 3, 'left'),
                 '-': Operator('-', 3, 'left'), '&': Operator('&', 2, 'left'), '=': Operator('=', 1, 'left'),
                 '<': Operator('<', 1, 'left'), '>': Operator('>', 1, 'left'), '<=': Operator('<=', 1, 'left'),
                 '>=': Operator('>=', 1, 'left'), '<>': Operator('<>', 1, 'left')}

    output = collections.deque()
    stack = []
    were_values = []
    arg_count = []

    for t in tokens:
        if t.ttype == "operand":

            output.append(create_node(t))
            if were_values:
                were_values.pop()
                were_values.append(True)

        elif t.ttype == "function":

            stack.append(t)
            arg_count.append(0)
            if were_values:
                were_values.pop()
                were_values.append(True)
            were_values.append(False)

        elif t.ttype == "argument":

            while stack and (stack[-1].tsubtype != "start"):
                output.append(create_node(stack.pop()))

            if were_values.pop():
                arg_count[-1] += 1
            were_values.append(False)

            if not len(stack):
                raise Exception("Mismatched or misplaced parentheses")

        elif t.ttype.startswith('operator'):

            if len(stack) > 0 and t.ttype.endswith('-prefix') and t.tvalue == "-":
                o1 = operators['u-']
            else:
                o1 = operators[t.tvalue]

            while stack and stack[-1].ttype.startswith('operator'):

                if stack[-1].ttype.endswith('-prefix') and stack[-1].tvalue == "-":
                    o2 = operators['u-']
                else:
                    o2 = operators[stack[-1].tvalue]

                if ((o1.associativity == "left" and o1.precedence <= o2.precedence)
                    or
                        (o1.associativity == "right" and o1.precedence < o2.precedence)):

                    output.append(create_node(stack.pop()))
                else:
                    break

            stack.append(t)

        elif t.tsubtype == "start":
            stack.append(t)

        elif t.tsubtype == "stop":

            while stack and stack[-1].tsubtype != "start":
                output.append(create_node(stack.pop()))

            if not stack:
                raise Exception("Mismatched or misplaced parentheses")

            stack.pop()

            if stack and stack[-1].ttype == "function":
                f = create_node(stack.pop())
                a = arg_count.pop()
                w = were_values.pop()
                if w:
                    a += 1
                f.num_args = a
                # print f, "has ",a," args"
                output.append(f)

    while stack:
        if stack[-1].tsubtype == "start" or stack[-1].tsubtype == "stop":
            raise Exception("Mismatched or misplaced parentheses")

        output.append(create_node(stack.pop()))

    # print "Stack is: ", "|".join(stack)
    # print "Ouput is: ", "|".join([x.tvalue for x in output])

    # convert to list
    result = [x for x in output]
    return result

def build_ast(expression):
    """build an AST from an Excel formula expression in reverse polish notation"""

    # use a directed graph to store the tree
    g = DiGraph()

    stack = []

    for n in expression:
        # Since the graph does not maintain the order of adding nodes/edges
        # add an extra attribute 'pos' so we can always sort to the correct order
        if isinstance(n, OperatorNode):
            if n.ttype == "operator-infix":
                arg2 = stack.pop()
                arg1 = stack.pop()
                g.add_node(arg1, {'pos': 1})
                g.add_node(arg2, {'pos': 2})
                g.add_edge(arg1, n)
                g.add_edge(arg2, n)
            else:
                arg1 = stack.pop()
                g.add_node(arg1, {'pos': 1})
                g.add_edge(arg1, n)

        elif isinstance(n, FunctionNode):
            args = [stack.pop() for _ in range(n.num_args)]
            args.reverse()
            for i, a in enumerate(args):
                g.add_node(a, {'pos': i})
                g.add_edge(a, n)
                # for i in range(n.num_args):
                #    G.add_edge(stack.pop(),n)
        else:
            g.add_node(n, {'pos': 0})

        stack.append(n)

    if len(stack) < 1:
        raise Exception("Error building ast, expression: [%s], graph: [%s]" % (expression, g))

    return g, stack.pop()

class Context(object):
    """A small context object that nodes in the AST can use to emit code"""

    def __init__(self, curcell, excel):
        # the current cell for which we are generating code
        self.curcell = curcell
        # a handle to an excel instance
        self.excel = excel

def create_node(t):
    """Simple factory function"""
    if t.ttype == "operand":
        if t.tsubtype == "range":
            return RangeNode(t)
        else:
            return OperandNode(t)
    elif t.ttype == "function":
        return FunctionNode(t)
    elif t.ttype.startswith("operator"):
        return OperatorNode(t)
    else:
        return ASTNode(t)

class Operator:
    """Small wrapper class to manage operators during shunting yard"""

    def __init__(self, value, precedence, associativity):
        self.value = value
        self.precedence = precedence
        self.associativity = associativity

class Spreadsheet(object):
    def __init__(self, g, cellmap):
        super(Spreadsheet, self).__init__()
        self.G = g
        self.cellmap = cellmap
        self.params = None

    def contains(self, cell):
        return cell in self.cellmap

    def plot_graph(self):
        import matplotlib.pyplot as plt

        pos = nx.spring_layout(self.G, iterations=2000)
        # pos=nx.spectral_layout(G)
        # pos = nx.random_layout(G)
        nx.draw_networkx_nodes(self.G, pos)
        nx.draw_networkx_edges(self.G, pos, arrows=True)
        nx.draw_networkx_labels(self.G, pos)
        plt.show()

    def set_value(self, cell, val, is_addr=True):
        if is_addr:
            cell = self.cellmap[cell]
            if cell.value != val:
                # We should always reset the successors when the value is changes, also when it's changed from a None value
                # to a not-None value, else the others aren't informed, and will not reset it's value such that the old
                # value will be used when evaluated.

                # TODOReset: enable when reset works correctly
                # self.reset_all(cell, True)
                cell.value = val
                # Because we set a value we indicate it as not-dirty. Suppose the cell contains a formula, like a
                # reference to another cell, setting the value will overwritte the behavior at that moment. When we "read"
                # the value through evaluation, it would (if not indicating it as not-dirty) evaluate the formula as
                # it's dirty through the reset call in  this method, such that it will set/store the value of the
                # reference and overwrite the set value "manually" by calling this method (as if it wasn't set)
                # Note: till it will be reset, such that it becomes dirty, the set value will be returned. When it's dirty
                # will be be evaluated that will result in using the referenced value (again)
                cell.dirty = False

    def reset(self, cell):
        self.reset_all(cell, cell.value is not None or isinstance(cell, CellRange))

    def reset_all(self, cell, reset_successors):
        if cell.reset_inprogress:
            # TODOReset: Reset already in progress, that easily happens due to the current reset implementation. See
            # the redmine issue to improve this.
            return
            # raise Exception("Cell reset in progress: [%s]" % (cell))
        try:
            cell.reset_inprogress = True
            # Note: we split the reset method, as before it would only perform the reset in case the cell value isn't None
            # already. However, that was restricted in case it was already None, such that a change in value would not reset
            # the successors such that the successors would use the old value.
            # CellRange instances are skipped as their value isn't used
            if (isinstance(cell, Cell)):
                cell.value = None

            # If already dirty, no need to continue, we assume that it's successors been made dirty as well.
            # The value of CellRange isn't used so we have to skip it
            if not cell.dirty or isinstance(cell, CellRange):
                if (isinstance(cell, Cell)):
                    cell.dirty = True
                    # TODOReset: Currently we don't reset ranges, as it will cause too many reset actions.
                    # It should be done when the reset issue is solved
                    # This seems to result in enough reset, as the unit test seem to work when changing values
                    if (reset_successors):
                        list(map(self.reset, self.G.successors_iter(cell)))
        finally:
            cell.reset_inprogress = False

    def print_value_tree(self, addr, indent):
        cell = self.cellmap[addr]
        print(("%s %s = %s" % (" " * indent, addr, cell.value)))
        for c in self.G.predecessors_iter(cell):
            self.print_value_tree(c.address(), indent + 1)

    # def recalculate(self, addr):
    #     for c in list(self.cellmap.values()):
    #         if isinstance(c, CellRange):
    #             self.evaluate_range(c, is_addr=False)
    #         else:
    #             self.evaluate(c, is_addr=False)
    #         return self.evaluate(c)

    # def evaluate_range(self, rng, is_addr=True):
    #     if is_addr:
    #         rng = self.cellmap[rng]
    #
    #     # its important that [] gets treated as false here
    #     if rng.value:
    #         return rng.value
    #     elif isinstance(rng.value, str):
    #         return rng.value
    #
    #     cells, nrows, ncols = rng.celladdrs, rng.nrows, rng.ncols
    #     if 1 == nrows or 1 == ncols:
    #         data = [self.evaluate(c) for c in cells]
    #     else:
    #         data = [[self.evaluate(c) for c in cells[i]] for i in range(len(cells))]
    #
    #     # Ensure to indicate the CellRange instance isn't dirty anymore, else it will get ignored next time such that
    #     # it's successors aren't reset and return the wrong value (example: B1(Cell)-> B1:B2(CellRange)-> Vlookup(Cell)
    #     rng.value = data
    #     rng.dirty = False
    #     return data

    def evaluate(self, cell, is_addr=True):
        if is_addr:
            cell = self.cellmap[cell]

            if cell.eval_inprogress:
                raise Exception("Evaluation of cell already in progress, [%s] for [%s], [%s]"
                                % (cell.address(), cell.python_expression, cell))

            # We only evaluate cells that have a formula and are dirty. In case they aren't dirty we return their
            # current value. We don't check if the value isn't None, as None might be a valid result of a formula
            # which can cause infinite, or unnecessary loops

            # TODOReset: enable when reset is solved. Below will result in every cell being eavaluated.
            # if not cell.formula or not cell.dirty:
            #     return cell.value
            if not cell.formula:
                return cell.value

        # TODOEd: is this function still needed,
        def eval_cell(address):
            # return address;
            try:
                # Note: we have to evaluate the cell and return the value, else an expression like "A3+A4" will fail
                # as python will try to "add" strings. This expression is processed by python directly and not through
                # a function in the excel_lib
                value = self.evaluate(address)
                if not value:
                    try:
                        if not value and cell.formula:
                            # The split will fail in case it's not a valid cell address
                            split_address(cell.formula[1:])
                            # In case it concerns a cell reference, it should return 0 in case the references cell is None
                            value = 0
                    except Exception:
                        pass
                return value
            # return self.evaluate(address)
            except:
                raise self.create_exception("Problem evaluating cell [%s]", address)

        # TODOEd: remove the eval_range calls? They aren't needed anymore, is replaced to the functions
        def eval_range(rng):
            try:
                # Changed: we return the range and let the "user" (function) decide when to evaluate.
                # Example: in case of a (v/h)lookup function we don't want to evaluate the complete search
                # range before feeding it to the lookup function, as the lookup function only use small parts
                # of the search rang. And evaluating a huge range can easily lead to circular references.
                return rng;
                # return self.evaluate_range(rng)
            except:
                raise self.create_exception("Problem evaluating cell [%s]", rng)

        try:
            cell.eval_inprogress = True
            print(("Evaluating: %s, %s" % (cell.address(), cell.python_expression)))
            if is_eval_allowed(self, cell):
                # indicate it's not dirty anymore before evaluation occurs to overcome infinite recursive loops.
                cell.dirty = False
                vv = eval(cell.compiled_expression)
            else:
                vv = cell.compiled_expression

            cell.dirty = False
            cell.value = vv
        except Exception as ex:
            if str(ex).startswith("Problem evaluating"):
                raise ex
            else:
                raise Exception(
                    "Problem evaluating: [%s] for [%s], [%s]" % (ex, cell.address(), cell.python_expression))
        finally:
            cell.eval_inprogress = False

        return cell.value

    def create_exception(self, msg, rng):
        if rng in self.cellmap:
            rng = self.cellmap[rng]
        return Exception(msg % rng)



# ExcelOpxWrapper : Can be run anywhere but only with post 2010 Excel formats
# noinspection PyBroadException
try:
    import numpy as np
except:
    pass

from openpyxl import load_workbook
from openpyxl.utils import range_boundaries
from .excel_util import ERR_NA, ERR_VALUE, ERR_REF, ERR_NAME, ERR_DIV
# from openpyxl.cell import Cell


import os
from os import path

import abc
from abc import abstractmethod

class ExcelWrapper(object, metaclass=abc.ABCMeta):
    @abstractmethod
    def rangednames(self):
        return

    @abstractmethod
    def connect(self):
        return

    @abstractmethod
    def save(self):
        return

    @abstractmethod
    def save_as(self, filename, delete_existing=False):
        return

    @abstractmethod
    def close(self):
        return

    @abstractmethod
    def quit(self):
        return

    @abstractmethod
    def set_sheet(self, s):
        return

    @abstractmethod
    def get_sheet(self):
        return

    @abstractmethod
    def get_range(self, range):
        return

    @abstractmethod
    def get_used_range(self):
        return

    @abstractmethod
    def get_active_sheet(self):
        return

    @abstractmethod
    def get_cell(self, r, c):
        return

    def get_value(self, r, c):
        return self.get_cell(r, c).value

    def set_value(self, r, c, val):
        self.get_cell(r, c).value = val

    def get_formula(self, r, c):
        f = self.get_cell(r, c).Formula
        return f if f.startswith("=") else None

    def has_formula(self, range):
        f = self.get_range(range).Formula
        if type(f) == str:
            return f and f.startswith("=")
        else:
            for t in f:
                if t[0].startswith("="):
                    return True
            return False

    def get_formula_from_range(self, range):
        f = self.get_range(range).Formula
        if isinstance(f, (list, tuple)):
            if any([x for x in f if x[0].startswith("=")]):
                return [x[0] for x in f]
            else:
                return None
        else:
            return f if f.startswith("=") else None

    def get_formula_or_value(self, name):
        r = self.get_range(name)
        return r.Formula or r.Value

    @abstractmethod
    def get_row(self, row):
        """"""
        return

    @abstractmethod
    def set_calc_mode(self, automatic=True):
        """"""
        return

    @abstractmethod
    def set_screen_updating(self, update):
        """"""
        return

    @abstractmethod
    def run_macro(self, macro):
        """"""
        return
# Excel range wrapper that distribute reduced api used by compiler (Formula & Value)


class OpxRange(object):
    def __init__(self, cells, cells_do):

        super(OpxRange, self).__init__()

        self.cells = cells
        self.cellsDO = cells_do

        # Added for caching as the function Formula is used many times, and I don't think the cells content changes.
        self.formulas = None
        self.values = None

    def reset(self):
        self.formulas = None
        self.values = None

    @property
    def Formula(self):
        if not self.formulas:
            self.formulas = ()
            for row in self.cells:
                col = ()
                for cell in row:
                    col += (str(cell.value),)
                self.formulas += (col,)
            if sum(map(len, self.formulas)) == 1:
                self.formulas = self.formulas[0][0]

        return self.formulas

    @property
    def value(self):
        if not self.values:
            self.values = []
            for row in self.cellsDO:
                col = ()
                for cell in row:
                    if cell.data_type == 'f':
                        col += (cell.value,)
                    elif cell.data_type == 'n':
                        col += (cell.value,)
                    elif cell.data_type == 's':
                        col += (cell.value,)
                    elif cell.data_type == 'b':
                        col += (cell.value,)
                    # we assume it concerns a formula with an error as it's input isn't correct yet.
                    elif cell.data_type == 'e' and cell.value and cell.value == ERR_NA:
                        # It's better not to treat the #N/A value as None as it might result in unncessary evaluations
                        col += (ERR_NA,)
                    elif cell.data_type == 'e' and cell.value and cell.value == ERR_VALUE:
                        col += (ERR_VALUE,)
                    elif cell.data_type == 'e' and cell.value and cell.value == ERR_REF:
                        col += (ERR_REF,)
                    elif cell.data_type == 'e' and cell.value and cell.value == ERR_NAME:
                        col += (ERR_NAME,)
                    elif cell.data_type == 'e' and cell.value and cell.value == ERR_DIV:
                        col += (ERR_DIV,)
                    # We raise an exception as it might result in strange errors:
                    # It's highly possible the dimension of the self.values doesn't correspond to the dimension of
                    # that of the self.formulas such that we could match a formula with the wrong value
                    else:  # We always add a value, else the dimensions of the values
                        raise Exception(
                            "Unknown cell data type: [%s], value: [%s], cell: [%s]. Ensure to support it to overcome dimension mismatch with formulas" % (
                            cell.data_type, cell.value, cell))
                        # col += None

                self.values += (col,)
            if sum(map(len, self.values)) == 1:
                self.values = self.values[0][0]

        return self.values

# OpenPyXl implementation for ExcelWrapper interface
class ExcelOpxWrapper(ExcelWrapper):
    def __init__(self, filename, ):

        super(ExcelWrapper, self).__init__()

        self.filename = path.abspath(filename)

    @property
    def rangednames(self):

        if self.workbook is None:
            return None

        rangename = []
        for named_range in self.workbook.get_named_ranges():
            for worksheet, range_alias in named_range.destinations:
                tuple_name = (len(rangename) + 1, str(named_range.name), str(worksheet.title + '!' + range_alias))
                rangename.append([tuple_name])
        return rangename

    def connect(self):
        self.workbook = load_workbook(self.filename)
        self.workbookDO = load_workbook(self.filename, data_only=True)

    def save(self):
        self.workbook.save(self.filename)

    def save_as(self, filename, delete_existing=False):
        if delete_existing and os.path.exists(filename):
            os.remove(filename)
        self.workbook.save(filename)

    def close(self):
        return

    def quit(self):
        return

    def set_sheet(self, s):
        self.workbook.active = self.workbook.index(self.workbook[s])
        self.workbookDO.active = self.workbookDO.index(self.workbookDO[s])

        # Removed, as it's deprecated.
        # self.workbook.active = self.workbook.get_index(self.workbook[s])
        # self.workbookDO.active = self.workbookDO.get_index(self.workbookDO[s])
        return self.workbook.active

    def get_sheet(self):
        return self.workbook.active

    def max_rows(self):
        return self.workbook.active.max_row

    def max_columns(self):
        return self.workbook.active.max_column

    def get_range(self, address):
        sheet = self.workbook.active
        sheetDO = self.workbookDO.active
        if address.find('!') > 0:
            title, address = address.split('!')
            sheet = self.workbook[title]
            sheetDO = self.workbookDO[title]

        min_col, min_row, max_col, max_row = range_boundaries(address.upper())
        cells = [[cell for cell in row] for row in
                 sheet.iter_rows(min_row=min_row, max_row=max_row, min_col=min_col, max_col=max_col)]
        cellsDO = [[cell for cell in row] for row in
                   sheetDO.iter_rows(min_row=min_row, max_row=max_row, min_col=min_col, max_col=max_col)]

        # Removed, as it's deprecated.
        # cells = [[cell for cell in row] for row in sheet.iter_rows(address)]
        # cellsDO = [[cell for cell in row] for row in sheetDO.iter_rows(address)]

        return OpxRange(cells, cellsDO)

    def get_used_range(self):
        return self.workbook.active.iter_rows()

    def get_active_sheet(self):
        return self.workbook.active.title

    def get_cell(self, r, c):
        # this could be improved in order not to call get_range
        return self.get_range(self.workbook.active.cell(None, r, c).coordinate)

    def get_row(self, row):
        return [self.get_value(row, col + 1) for col in range(self.workbook.active.max_column)]

    def set_calc_mode(self, automatic=True):
        raise Exception('Not implemented')

    def set_screen_updating(self, update):
        raise Exception('Not implemented')

    def run_macro(self, macro):
        raise Exception('Not implemented')
